from argparse import ArgumentParser

import yaml
import schema
from schema import SchemaError, Schema, Optional

from ccpi.viewer.utils.io import ImageReader, ImageWriter
'''
This command line tool takes a dataset file and a yaml file as input.
It resamples or crops the dataset as it reads it in, and then writes
out the resulting dataset to a file.

Supported file types for reading:
hdf5, nxs, mha, raw, numpy

Supported file types for writing:
hdf5, nxs, mha

'''

EXAMPLE_YAML_FILE = '''
input:
    file_name: '24737_fd_normalised.nxs'
    shape: (1024,1024,1024)
    is_fortran: False
    is_big_endian: True
    typecode: 'float32'
    dataset_name: '/entry1/tomo_entry/data/data' # valid for HDF5 and Zarr
resample:
    target_size: 1
    resample_z: True
output:
    file_name: 'this_fname.nxs'
    format: 'hdf5' # npy, METAImage, NIFTI (or Zarr to come)
'''
'''

THIS IS WHAT THE OUTPUT STRUCTURE OF AN EXAMPLE OUTPUT HDF5 FILE WOULD LOOK LIKE:
Entry 1 contains attributes of the original dataset
Entry 2 contains the resampled dataset and attributes

- entry1 : <HDF5 group "/entry1" (1 members)>
            - tomo_entry : <HDF5 group "/entry1/tomo_entry" (1 members)>
                    - data : <HDF5 group "/entry1/tomo_entry/data" (1 members)>
                            - data : <HDF5 dataset "data": shape (500, 100, 600), type "<f4">
                                            - bit_depth : 64
                                            - file_name : test_3D_data.npy
                                            - header_length : 128
                                            - is_big_endian : False
                                            - origin : [0. 0. 0.]
                                            - shape : [500 100 600]
                                            - spacing : [1 1 1]
                                            - resampled: False
                                            - cropped: False
    - entry2 : <HDF5 group "/entry2" (1 members)>
            - tomo_entry : <HDF5 group "/entry2/tomo_entry" (1 members)>
                    - data : <HDF5 group "/entry2/tomo_entry/data" (1 members)>
                            - data : <HDF5 dataset "data": shape (500, 18, 109), type "<f8">
                                            - cropped : False
                                            - origin : [2.23861279 2.23861279 0.        ]
                                            - resample_z : True
                                            - resampled : True
                                            - spacing : [5.47722558 5.47722558 1.        ]
                                            - original_dataset: /entry1/tomo_entry/data/data
'''

# This validates the input yaml file:
schema = Schema({
    'input': {
        'file_name': str,
        Optional('shape'): tuple,  # only for raw
        Optional('is_fortran'): bool,  # only for raw
        Optional('is_big_endian'): bool,  # only for raw
        Optional('typecode'): str,  # only for raw
        Optional('dataset_name'): str
    },  # only for hdf5 # need to set default
    'resample': {
        'target_size': float,
        'resample_z': bool
    },
    'output': {
        'file_name': str,
        'format': str
    }
})


def parse_arguments():
    parser = ArgumentParser(prog='resampler',
                            description='Resamples a dataset file & writes out to a file.' +
                            ' Either specify a yaml file or set at minimum: -i, -o and -target_size')

    parser.add_argument(
        '-f',
        help='Input yaml file. May be used in place of all other arguments. If set, all other arguments are ignored.',
        type=str)
    parser.add_argument('--example', help='Prints an example input yaml file.', action='store_true')

    parser.add_argument('-i', help='Input dataset filename. Required if -f is not set.')
    parser.add_argument('--dataset_name',
                        help='Dataset name, only required if input file is HDF5/Nexus format.',
                        type=str,
                        default='/entry1/tomo_entry/data/data')
    parser.add_argument('--shape',
                        help='Shape of input dataset file, only required if input file is raw format.',
                        type=lambda s: [int(item) for item in s.strip('[').strip(']').split(',')])
    parser.add_argument('--is_fortran',
                        help='Whether input dataset file is fortran order, only required if input file is raw format.')
    parser.add_argument('--is_big_endian',
                        help='Whether input dataset file is big endian, only required if input file is raw format.')
    parser.add_argument('--typecode',
                        help='Typecode of input dataset, only required if input file is raw format.',
                        type=str)

    parser.add_argument('-target_size',
                        help='Target size to downsample dataset to, in MB. Required if -f is not set.',
                        type=float)
    parser.add_argument('--resample_z',
                        help='Whether to resample along the z axis of the dataset. Optional.',
                        default=True)

    parser.add_argument('-o', help='Output filename. Required if -f is not set.')
    parser.add_argument('--out_format',
                        help='File format to write downsampled data to',
                        choices=['hdf5', 'nxs', 'mha'],
                        type=str,
                        default='nxs')

    args = parser.parse_args()

    return args


def get_params_from_args(args):

    if args.f is not None:

        with open(args.f) as f:
            params_raw = yaml.safe_load(f)
            params = {}
            for key, dict in params_raw.items():
                # each of the values in data_raw is a dict
                params[key] = {}
                for sub_key, value in dict.items():
                    try:
                        params[key][sub_key] = eval(value)
                    except:
                        params[key][sub_key] = value
        try:
            schema.validate(params)
        except SchemaError as e:
            print(e)

    else:
        for a in [args.i, args.target_size, args.o]:
            if a is None:
                raise Exception("If yaml file is not set: -i, --target_size, and -o must be set")

        params = {
            'input': {
                'file_name': args.i
            },
            'resample': {
                'target_size': args.target_size
            },
            'output': {
                'file_name': args.o
            }
        }

        args_input = ['shape', 'typecode', 'dataset_name']
        bool_args_input = ['is_fortran', 'is_big_endian']

        for a in args_input:
            if eval(f"args.{a}") is not None:
                params['input'][a] = eval(f"args.{a}")

        for a in bool_args_input:
            if eval(f"args.{a}") is not None:
                params['input'][a] = eval(eval(f"args.{a}"))

        if args.resample_z is not None:
            params['resample']['resample_z'] = eval(args.resample_z)

        if args.out_format is not None:
            params['output']['format'] = args.out_format

    return params


def main():
    args = parse_arguments()
    # Do nothing else if we are just printing an example yaml file:
    if args.example:
        print(EXAMPLE_YAML_FILE)
        return
    params = get_params_from_args(args)

    raw_attrs = None
    dataset_name = None
    if 'input' in params.keys():
        raw_attrs = {}
        for key, value in params['input'].items():
            if key == 'dataset_name':
                dataset_name = value
            elif key == 'file_name':
                pass
            else:
                raw_attrs[key] = value

    target_size = params['resample']['target_size']
    if target_size is not None:
        target_size *= 1e6

    reader = ImageReader(file_name=params['input']['file_name'],
                         resample=True,
                         target_size=target_size,
                         resample_z=params['resample']['resample_z'],
                         raw_image_attrs=raw_attrs,
                         hdf5_dataset_name=dataset_name)
    downsampled_image = reader.Read()
    original_image_attrs = reader.GetOriginalImageAttrs()
    loaded_image_attrs = reader.GetLoadedImageAttrs()

    writer = ImageWriter()
    writer.SetFileName(params['output']['file_name'])
    writer.SetFileFormat(params['output']['format'])
    writer.SetOriginalDataset(None, original_image_attrs)
    writer.AddChildDataset(downsampled_image, loaded_image_attrs)
    writer.Write()


if __name__ == '__main__':
    main()



#
#   Copyright 2022 STFC, United Kingdom Research and Innovation
#
#   Author 2022 Samuel Jones
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import unittest

import vtk

from ccpi.viewer.utils import CameraData


class CameraDataTest(unittest.TestCase):

    def setUp(self):
        self.camera = vtk.vtkCamera()
        self.cam_pos = (1., 1., 1.)
        self.focal_pos = (2., 2., 2.)
        self.view_up = (0.57735026, 0.57735026, 0.57735026)
        self.camera.SetPosition(*self.cam_pos)
        self.camera.SetFocalPoint(*self.focal_pos)
        self.camera.SetViewUp(*self.view_up)

        self.data = CameraData(self.camera)

    def test_init_sets_all_values(self):
        self.assertEqual(self.data.position, self.cam_pos)
        self.assertEqual(self.data.focalPoint, self.focal_pos)
        self.assertAlmostEqual(self.data.viewUp[0], self.view_up[0])
        self.assertAlmostEqual(self.data.viewUp[1], self.view_up[1])
        self.assertAlmostEqual(self.data.viewUp[2], self.view_up[2])

    def test_copy_data_to_camera_does_so(self):
        camera_to_copy_to = vtk.vtkCamera()

        self.assertNotEqual(self.cam_pos, camera_to_copy_to.GetPosition())
        self.assertNotEqual(self.focal_pos, camera_to_copy_to.GetFocalPoint())
        self.assertNotEqual(self.view_up, camera_to_copy_to.GetViewUp())

        CameraData.CopyDataToCamera(self.data, camera_to_copy_to)

        self.assertEqual(self.cam_pos, camera_to_copy_to.GetPosition())
        self.assertEqual(self.focal_pos, camera_to_copy_to.GetFocalPoint())
        self.assertAlmostEqual(self.view_up[0], camera_to_copy_to.GetViewUp()[0])
        self.assertAlmostEqual(self.view_up[1], camera_to_copy_to.GetViewUp()[1])
        self.assertAlmostEqual(self.view_up[2], camera_to_copy_to.GetViewUp()[2])



import vtk
from ccpi.viewer import (ALT_KEY, CONTROL_KEY, CROSSHAIR_ACTOR, CURSOR_ACTOR, HELP_ACTOR, HISTOGRAM_ACTOR,
                         LINEPLOT_ACTOR, OVERLAY_ACTOR, SHIFT_KEY, SLICE_ACTOR, SLICE_ORIENTATION_XY,
                         SLICE_ORIENTATION_XZ, SLICE_ORIENTATION_YZ)
from ccpi.viewer.utils.io import SaveRenderToPNG


class ViewerEventManager(object):

    def __init__(self):
        # If all values are false it signifies no event
        self.events = {
            "PICK_EVENT": False,  # left  mouse
            "WINDOW_LEVEL_EVENT": False,  # alt + right mouse + move
            "ZOOM_EVENT": False,  # shift + right mouse + move
            "PAN_EVENT": False,  # ctrl + right mouse + move
            "CREATE_ROI_EVENT": False,  # ctrl + left mouse
            "DELETE_ROI_EVENT": False,  # alt + left mouse
            "SHOW_LINE_PROFILE_EVENT": False,  # l
            "UPDATE_WINDOW_LEVEL_UNDER_CURSOR": False,  # Mouse move + w
            "RECTILINEAR_WIPE": False  # activated by key 2, updates by mouse move
        }

    def __str__(self):
        return str(self.events)

    def On(self, event):
        self.events[event] = True

    def Off(self, event):
        self.events[event] = False

    def setAllInactive(self):
        self.events = {x: False for x in self.events}

    def isActive(self, event):
        return self.events[event]

    def isAllInactive(self):
        """Returns True if all events are inactive"""
        return all(not x for x in self.events.values())


class CILViewerBase():
    '''
    Base Class for CILViewers.

    When making a subclass, the following must be set in the __init__:

    self.setInteractorStyle(style)

    '''

    def __init__(self, dimx=600, dimy=600, renWin=None, iren=None, ren=None, debug=False):
        # Handle arguments:
        # create a renderer
        if ren is None:
            ren = vtk.vtkRenderer()
        self.ren = ren

        # create a rendering window
        if renWin is None:
            renWin = vtk.vtkRenderWindow()

        renWin.SetSize(dimx, dimy)
        renWin.AddRenderer(self.ren)
        self.renWin = renWin

        if iren is not None:
            self.iren = iren
        else:
            self.iren = vtk.vtkRenderWindowInteractor()

        self.iren.SetRenderWindow(self.renWin)

        self.ren.SetBackground(.1, .2, .4)

        # img 3D as slice
        self.img3D = None
        self.slicenos = [0, 0, 0]
        self.sliceOrientation = SLICE_ORIENTATION_XY

        imageSlice = vtk.vtkImageSlice()
        imageSliceMapper = vtk.vtkImageSliceMapper()
        imageSlice.SetMapper(imageSliceMapper)
        imageSlice.GetProperty().SetInterpolationTypeToNearest()
        self.imageSlice = imageSlice
        self.imageSliceMapper = imageSliceMapper

        self.voi = vtk.vtkExtractVOI()
        self.ia = vtk.vtkImageHistogramStatistics()
        self.ia.SetAutoRangePercentiles(5.0, 95.)

        self.helpActor = vtk.vtkActor2D()
        self.helpActor.GetPositionCoordinate().SetCoordinateSystemToNormalizedDisplay()
        self.helpActor.GetPositionCoordinate().SetValue(0.1, 0.5)
        self.helpActor.VisibilityOff()

        # axis orientation widget
        om = vtk.vtkAxesActor()
        ori = vtk.vtkOrientationMarkerWidget()
        ori.SetOutlineColor(0.9300, 0.5700, 0.1300)
        ori.SetInteractor(self.iren)
        ori.SetOrientationMarker(om)
        ori.SetViewport(0.0, 0.0, 0.4, 0.4)
        ori.SetEnabled(1)
        ori.InteractiveOff()
        self.orientation_marker = ori
        # axes labels
        self.axisLabelsText = self.getCurrentAxisLabelsText()

        # holder for list of actors and widgets
        self.actors = {}
        self.widgets = {}

        #initial Window/Level
        self.InitialLevel = 0
        self.InitialWindow = 0

        #ViewerEvent
        self.event = ViewerEventManager()

        self.histogramPlotActor = vtk.vtkXYPlotActor()
        self.histogramPlotActor.ExchangeAxesOff()
        self.histogramPlotActor.SetXLabelFormat("%g")
        self.histogramPlotActor.SetXLabelFormat("%g")
        self.histogramPlotActor.SetAdjustXLabels(3)
        self.histogramPlotActor.SetXTitle("Level")
        self.histogramPlotActor.SetYTitle("N")
        self.histogramPlotActor.SetXValuesToValue()
        self.histogramPlotActor.SetPlotColor(0, (0, 1, 1))

    def setInteractorStyle(self, style):
        self.style = style
        self.iren.SetInteractorStyle(self.style)
        self.iren.Initialize()

    def getInteractor(self):
        return self.iren

    def getRenderer(self):
        return self.ren

    def getRenderWindow(self):
        '''returns the render window'''
        return self.renWin

    def getCamera(self):
        '''returns the active camera'''
        return self.ren.GetActiveCamera()

    def startRenderLoop(self):
        self.iren.Start()

    def saveRender(self, filename, renWin=None):
        '''Save the render window to PNG file'''
        if renWin is None:
            renWin = self.renWin
        SaveRenderToPNG(self.renWin, filename)

    def validateValue(self, value, axis):
        dims = self.img3D.GetDimensions()
        max_slice = [x - 1 for x in dims]

        axis_int = {'x': SLICE_ORIENTATION_YZ, 'y': SLICE_ORIENTATION_XZ, 'z': SLICE_ORIENTATION_XY}

        if axis in axis_int.keys():
            i = axis_int[axis]
        else:
            raise KeyError

        if value < 0:
            return 0
        if value > max_slice[i]:
            return max_slice[i]
        else:
            return value

    # METHODS ON THE FULL 3D IMAGE DATA: ------------------------------------------------

    def setInput3DData(self, imageData):
        raise NotImplementedError("Implemented in the subclasses.")

    def getImageHistogramStatistics(self, method, slice=False):
        '''
        returns histogram statistics for either the image
        or gradient of the image depending on the method
        if slice = True, calculates for the slice instead of
        the entire image volume.
        '''
        ia = vtk.vtkImageHistogramStatistics()

        if slice:
            input_data = self.voi.GetOutput()
        else:
            input_data = self.img3D

        if method == 'scalar':
            ia.SetInputData(input_data)
        else:
            grad = vtk.vtkImageGradientMagnitude()
            grad.SetInputData(input_data)
            grad.SetDimensionality(3)
            grad.Update()
            ia.SetInputData(grad.GetOutput())
        ia.Update()
        return ia

    def getImageMapRange(self, percentiles, method):
        '''
        uses percentiles to generate min and max values in either
        the image or image gradient (depending on method) for which
        the colormap or opacity are displayed.
        '''
        ia = self.getImageHistogramStatistics(method)
        ia.SetAutoRangePercentiles(*percentiles)
        ia.Update()
        min, max = ia.GetAutoRange()
        return min, max

    def getImageMapWholeRange(self, method):
        '''
        Parameters
        -----------
        method: string : ['scalar', 'gradient']
            'scalar' - returns full range of values in the 3D image
            'gradient' - returns full range of values in 3D image's gradient
        '''

        ia = self.getImageHistogramStatistics(method)

        return ia.GetMinimum(), ia.GetMaximum()

    # METHODS ON THE SLICE: --------------------------------------------------------

    def getSliceOrientation(self):
        return self.sliceOrientation

    def setActiveSlice(self, sliceno):
        self.slicenos[self.getSliceOrientation()] = sliceno

    def getActiveSlice(self):
        return self.slicenos[self.getSliceOrientation()]

    # Set interpolation on
    def setInterpolateOn(self):
        self.imageSlice.GetProperty().SetInterpolationTypeToLinear()
        self.renWin.Render()

    # Set interpolation off
    def setInterpolateOff(self):
        self.imageSlice.GetProperty()\
            .SetInterpolationTypeToNearest()
        self.renWin.Render()

    def setSliceColorWindowLevel(self, window, level):
        '''
        Set the window and level for the 2D slice of the 3D image.
        '''
        # Level is the average of min and max, and the window is the difference.
        self.imageSlice.GetProperty().SetColorLevel(level)
        self.imageSlice.GetProperty().SetColorWindow(window)
        self.imageSlice.Update()
        self.ren.Render()
        self.renWin.Render()

    def setSliceColorPercentiles(self, min_percentage, max_percentage):
        min_val, max_val = self.getSliceMapRange((min_percentage, max_percentage), 'scalar')
        self.setSliceMapRange(min_val, max_val)

    def getSliceColorPercentiles(self):
        min_val, max_val = self.getSliceMapWholeRange('scalar')
        min_color, max_color = self.getSliceMapRange()
        min_percentage = (min_color - min_val) / (max_val - min_val) * 100
        max_percentage = (max_color - min_val) / (max_val - min_val) * 100
        return min_percentage, max_percentage

    def setSliceColorWindow(self, window):
        '''
        Set the window for the 2D slice of the 3D image.
        '''
        self.imageSlice.GetProperty().SetColorWindow(window)
        self.imageSlice.Update()
        self.ren.Render()
        self.renWin.Render()

    def setSliceColorLevel(self, level):
        '''
        Set the level for the 2D slice of the 3D image.
        '''
        self.imageSlice.GetProperty().SetColorLevel(level)
        self.imageSlice.Update()
        self.ren.Render()
        self.renWin.Render()

    def getSliceColorWindow(self):
        '''
        Get the window for the 2D slice of the 3D image.
        '''
        return self.imageSlice.GetProperty().GetColorWindow()

    def getSliceColorLevel(self):
        '''
        Get the level for the 2D slice of the 3D image.
        '''
        return self.imageSlice.GetProperty().GetColorLevel()

    def getSliceWindowLevelFromRange(self, cmin, cmax):
        # set the level to the average between the percentiles
        level = (cmin + cmax) / 2
        # accommodates all values between the level and the percentiles
        window = cmax - cmin

        return window, level

    def getSliceMapWholeRange(self, method):
        '''
        Parameters
        -----------
        method: string : ['scalar', 'gradient']
            'scalar' - returns full range of values in image slice
            'gradient' - returns full range of values in image slice's gradient
        '''

        return self.getSliceMapRange((0, 100), method)

    def getSliceMapRange(self, percentiles, method='scalar'):
        '''
        uses percentiles to generate min and max values in
        the 2D slice of the 3D image, for which
        the colormap is displayed.
        '''

        ia = self.getImageHistogramStatistics(method, slice=True)
        ia.SetAutoRangePercentiles(*percentiles)
        ia.Update()
        min, max = ia.GetAutoRange()
        return min, max

    def setSliceMapRange(self, min, max):
        '''
        Parameters
        -----------
        min, max: float, default: the raw value of the 80. percentile for min, and the raw value of the 99. percentile for max.
            the upper and lower image values that the 
            color will be mapped to.
        update_pipeline: bool
            whether to immediately update the pipeline with this new
            setting
        '''
        window, level = self.getSliceWindowLevelFromRange(min, max)
        self.setSliceColorLevel(level)
        self.setSliceColorWindow(window)

    def autoWindowLevelOnSliceRange(self, update_slice=True):
        '''Auto-adjusts window-level for the slice, based on the 5 and 95th percentiles of the current slice.'''
        self.ia.SetAutoRangePercentiles(5.0, 95.)
        cmin, cmax = self.ia.GetAutoRange()
        window, level = self.getSliceWindowLevelFromRange(cmin, cmax)

        self.imageSlice.GetProperty().SetColorLevel(level)
        self.imageSlice.GetProperty().SetColorWindow(window)

        if update_slice:
            self.style.UpdateImageSlice()

    def addWidgetReference(self, widget, name):
        '''Adds widget to dictionary of widgets'''
        if self.getWidget(name) is not None:
            raise ValueError(f'Could not save reference to widget, as a widget with name {name} already exists.')
        self.widgets[name] = widget

    def getWidget(self, name):
        return self.widgets.get(name)

    def deleteWidget(self, name):
        ''' deletes a widget
        Parameters:
        name: string
            reference name given to the widget in the 
            dictionary of widgets'''

        widget = self.getWidget(name)

        if widget is not None:
            widget.Off()
            del widget
            self.widgets.pop(name)

    def setVisualisationDownsampling(self, value):
        self.visualisation_downsampling = value
        if value != [1, 1, 1]:
            self.image_is_downsampled = True
        else:
            self.image_is_downsampled = False

    def getVisualisationDownsampling(self):
        return self.visualisation_downsampling

    def getCurrentAxisLabelsText(self):
        '''Returns the current labels on the axis widget.'''
        om = self.orientation_marker.GetOrientationMarker()
        return [om.GetXAxisLabelText(), om.GetYAxisLabelText(), om.GetZAxisLabelText()]

    def setAxisLabels(self, labels=['x', 'y', 'z'], overwrite_flag=True):
        '''Sets the axes widget labels.
        
        Parameters
        ----------
        labels : list of str
        overwrite_flag : bool
            If True the attribute 'axisLabelText' is overwritten, if False it is not overwritten'''
        if type(labels) != list:
            raise TypeError("Labels must be a list of strings")
        if overwrite_flag is True:
            self.axisLabelsText = labels
        om = self.orientation_marker.GetOrientationMarker()
        om.SetXAxisLabelText(labels[0])
        om.SetYAxisLabelText(labels[1])
        om.SetZAxisLabelText(labels[2])



import os
import sys
import unittest
from unittest import mock

import vtk
from ccpi.viewer.CILViewer import CILViewer
from ccpi.viewer.CILViewer2D import CILViewer2D
from ccpi.viewer.ui.main_windows import ViewerMainWindow
from PySide2.QtWidgets import QApplication

# skip the tests on GitHub actions
if os.environ.get('CONDA_BUILD', '0') == '1':
    skip_as_conda_build = True
else:
    skip_as_conda_build = False

print("skip_as_conda_build is set to ", skip_as_conda_build)


@unittest.skipIf(skip_as_conda_build, "On conda builds do not do any test with interfaces")
class TestViewerMainWindow(unittest.TestCase):
    ''' Methods which have their full functionality tested:
    - onAppSettingsDialogAccepted
    - setDefaultDownsampledSize
    - getDefaulDownsampledSize
    - getTargetImageSize
    - updateViewerCoords
    '''

    def setUp(self):
        pass

    def tearDown(self) -> None:
        pass

    def test_all(self):
        # https://stackoverflow.com/questions/5387299/python-unittest-testcase-execution-order
        # https://stackoverflow.com/questions/11145583/unit-and-functional-testing-a-pyside-based-application
        self._test_init()
        self._test_init_calls_super_init()
        self._test_createAppSettingsDialog_calls_setAppSettingsDialogWidgets()
        self._test_acceptViewerSettings_when_gpu_checked()
        self._test_acceptViewerSettings_when_gpu_unchecked()
        self._test_setDefaultDownsampledSize()
        self._test_getDefaultDownsampledSize()
        self._test_getTargetImageSize_when_vis_size_is_None()
        self._test_getTargetImageSize_when_vis_size_is_not_None()
        self._test_updateViewerCoords_with_display_unsampled_coords_selected()
        self._test_updateViewerCoords_with_display_downsampled_coords_selected()
        self._test_updateViewerCoords_with_3D_viewer()
        self._test_updateViewerCoords_with_no_img3D()

    def _test_init(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        assert vmw is not None

    def _test_init_calls_super_init(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        # If the super init is called, then the following attributes should be set:
        assert vmw.app_name == "testing app name"
        assert vmw.threadpool is not None
        assert vmw.progress_windows == {}

    def _test_createAppSettingsDialog_calls_setAppSettingsDialogWidgets(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        vmw.setAppSettingsDialogWidgets = mock.MagicMock()
        vmw.setViewerSettingsDialogWidgets = mock.MagicMock()
        vmw.createAppSettingsDialog()
        vmw.setAppSettingsDialogWidgets.assert_called_once()
        vmw.setViewerSettingsDialogWidgets.assert_called_once()

    def _test_acceptViewerSettings_when_gpu_checked(self):

        vmw = self._setup_acceptViewerSettings_tests()

        vmw.acceptViewerSettings()

        vmw.settings.assert_has_calls(
            [mock.call.setValue('use_gpu_volume_mapper', True),
             mock.call.setValue('vis_size', 1.0)], any_order=True)

        assert isinstance(vmw.viewers[0].volume_mapper, vtk.vtkSmartVolumeMapper)

    def _test_acceptViewerSettings_when_gpu_unchecked(self):

        vmw, settings_dialog = self._setup_acceptViewerSettings_tests()

        vmw.acceptViewerSettings(settings_dialog)

        vmw.settings.assert_has_calls(
            [mock.call.setValue('use_gpu_volume_mapper', False),
             mock.call.setValue('vis_size', 1.0)], any_order=True)

        assert isinstance(vmw.viewers[0].volume_mapper, vtk.vtkFixedPointVolumeRayCastMapper)

    def _setup_acceptViewerSettings_tests(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        vmw.settings = mock.MagicMock()
        vmw.settings.setValue = mock.MagicMock()
        vis_size_field = mock.MagicMock()
        vis_size_field.value.return_value = 1.0
        gpu_checkbox_field = mock.MagicMock()
        gpu_checkbox_field.isChecked.return_value = True
        dark_checkbox_field = mock.MagicMock()
        dark_checkbox_field.isChecked.return_value = False
        settings_dialog = mock.MagicMock()
        settings_dialog.widgets = {
            'vis_size_field': vis_size_field,
            'gpu_checkbox_field': gpu_checkbox_field,
            'dark_checkbox_field': dark_checkbox_field
        }
        viewer3D = CILViewer()
        viewer3D.volume_mapper = None
        vmw._viewers = [viewer3D]
        vmw._vs_dialog = settings_dialog
        return vmw

    def _test_acceptViewerSettings_when_gpu_unchecked(self):
        vmw = self._setup_acceptViewerSettings_tests()
        vmw._vs_dialog.widgets['gpu_checkbox_field'].isChecked.return_value = False

        vmw.acceptViewerSettings()
        vmw.settings.assert_has_calls(
            [mock.call.setValue('use_gpu_volume_mapper', False),
             mock.call.setValue('vis_size', 1.0)], any_order=True)

        assert isinstance(vmw.viewers[0].volume_mapper, vtk.vtkFixedPointVolumeRayCastMapper)

    def _test_setDefaultDownsampledSize(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        vmw.setDefaultDownsampledSize(5)
        assert vmw.default_downsampled_size == 5

    def _test_getDefaultDownsampledSize(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        # Test what the default value is:
        assert vmw.getDefaultDownsampledSize() == 512**3
        vmw.default_downsampled_size = 5
        assert vmw.getDefaultDownsampledSize() == 5

    def _test_getTargetImageSize_when_vis_size_is_None(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        vmw.settings.setValue("vis_size", None)
        vmw.getDefaultDownsampledSize = mock.MagicMock()
        vmw.getDefaultDownsampledSize.return_value = 512**3
        returned_target_size = vmw.getTargetImageSize()
        vmw.getDefaultDownsampledSize.assert_called_once()
        assert (returned_target_size == 512**3)

    def _test_getTargetImageSize_when_vis_size_is_not_None(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        vmw.settings.setValue("vis_size", 5)
        vmw.getDefaultDownsampledSize = mock.MagicMock()
        returned_target_size = vmw.getTargetImageSize()
        vmw.getDefaultDownsampledSize.assert_not_called()
        assert (returned_target_size == 5 * (1024**3))

    def _test_updateViewerCoords_with_display_unsampled_coords_selected(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        viewer2D = CILViewer2D()
        viewer2D.visualisation_downsampling = [2, 2, 2]
        viewer2D.img3D = vtk.vtkImageData()
        viewer2D.setDisplayUnsampledCoordinates = mock.MagicMock()
        viewer2D.updatePipeline = mock.MagicMock()
        vmw.viewer_coords_dock.viewers = [viewer2D]
        viewer_coords_widgets = vmw.viewer_coords_dock.getWidgets()
        viewer_coords_widgets['coords_combo_field'].setCurrentIndex(0)
        viewer_coords_widgets['coords_warning_field'].setVisible = mock.MagicMock()
        vmw.updateViewerCoords()

        # Expect display unsampled coords to be called with True
        # and the warning field to be visible:
        viewer2D.setDisplayUnsampledCoordinates.assert_called_once_with(True)
        viewer_coords_widgets['coords_warning_field'].setVisible.assert_called_once_with(True)
        viewer2D.updatePipeline.assert_called_once()

        viewer2D.updatePipeline = mock.MagicMock()  # reset call count
        viewer2D.visualisation_downsampling = [1, 1, 1]
        vmw.updateViewerCoords()
        # Expect the warning field to be hidden:
        viewer_coords_widgets['coords_warning_field'].setVisible.assert_called_with(False)
        viewer2D.updatePipeline.assert_called_once()

    def _test_updateViewerCoords_with_display_downsampled_coords_selected(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        viewer2D = CILViewer2D()
        viewer2D.visualisation_downsampling = [2, 2, 2]
        viewer2D.img3D = vtk.vtkImageData()
        viewer2D.setDisplayUnsampledCoordinates = mock.MagicMock()
        viewer2D.updatePipeline = mock.MagicMock()
        vmw.viewer_coords_dock.viewers = [viewer2D]
        viewer_coords_widgets = vmw.viewer_coords_dock.getWidgets()
        viewer_coords_widgets['coords_combo_field'].setCurrentIndex(1)
        viewer_coords_widgets['coords_warning_field'].setVisible = mock.MagicMock()
        vmw.updateViewerCoords()

        # Expect display unsampled coords to be called with False
        # and the warning field to be hidden:
        viewer2D.setDisplayUnsampledCoordinates.assert_called_with(False)
        viewer_coords_widgets['coords_warning_field'].setVisible.assert_called_once_with(False)
        viewer2D.updatePipeline.assert_called()

    def _test_updateViewerCoords_with_3D_viewer(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        viewer3D = CILViewer()
        viewer3D.visualisation_downsampling = [2, 2, 2]
        viewer3D.updatePipeline = mock.MagicMock()
        viewer3D.setDisplayUnsampledCoordinates = mock.MagicMock()
        vmw.viewer_coords_dock.viewers = [viewer3D]
        viewer_coords_widgets = vmw.viewer_coords_dock.getWidgets()
        viewer_coords_widgets['coords_combo_field'].setCurrentIndex(1)
        viewer_coords_widgets['coords_warning_field'].setVisible = mock.MagicMock()
        vmw.updateViewerCoords()

        # Expect we don't call anything:
        viewer3D.setDisplayUnsampledCoordinates.assert_not_called()
        viewer_coords_widgets['coords_warning_field'].setVisible.assert_not_called()
        viewer3D.updatePipeline.assert_not_called()

    def _test_updateViewerCoords_with_no_img3D(self):
        vmw = ViewerMainWindow(title="Testing Title", app_name="testing app name")
        viewer2D = CILViewer2D()
        viewer2D.visualisation_downsampling = [2, 2, 2]
        viewer2D.img3D = None
        viewer2D.setDisplayUnsampledCoordinates = mock.MagicMock()
        viewer2D.updatePipeline = mock.MagicMock()
        vmw.viewer_coords_dock.viewers = [viewer2D]
        viewer_coords_widgets = vmw.viewer_coords_dock.getWidgets()
        viewer_coords_widgets['coords_combo_field'].setCurrentIndex(1)
        viewer_coords_widgets['coords_warning_field'].setVisible = mock.MagicMock()
        vmw.updateViewerCoords()

        # Expect we don't call anything:
        viewer2D.setDisplayUnsampledCoordinates.assert_not_called()
        viewer_coords_widgets['coords_warning_field'].setVisible.assert_not_called()
        viewer2D.updatePipeline.assert_not_called()


if __name__ == '__main__':
    unittest.main()
